FORMAT_CHARS = {
    u'int8': u'b',
    u'uint8': u'B',
    u'int16': u'h',
    u'uint16': u'H',
    u'int32': u'i',
    u'uint32': u'I',
    u'int64': u'q',
    u'uint64': u'Q',
    u'float32': u'f',
    u'float64': u'd',
}

ENDIANNESS = {
    u'little': u'<',
    u'big': u'>',
}
